import torch
import torch.nn as nn
import torch.nn.functional as F

from .layer_utils.block_diagonal import BlockDiagonal


class mLSTMblock(nn.Module):
    """MLSTM block with convolutions, gated mechanisms, and projection layers.

    Parameters
    ----------
    x_example : torch.Tensor
        Example input tensor for defining input dimensions.
    factor : float
        Factor to scale hidden size relative to input size.
    depth : int
        Depth of block diagonal layers.
    dropout : float, optional
        Dropout probability (default is 0.2).
    """

    def __init__(
        self,
        input_size,
        hidden_size,
        num_layers,
        bidirectional=None,
        batch_first=None,
        nonlinearity=F.silu,
        dropout=0.2,
        bias=True,
    ):
        super().__init__()
        self.input_size = input_size
        self.hidden_size = hidden_size
        self.activation = nonlinearity

        self.ln = nn.LayerNorm(self.input_size)

        self.left = nn.Linear(self.input_size, self.hidden_size)
        self.right = nn.Linear(self.input_size, self.hidden_size)

        self.conv = nn.Conv1d(
            in_channels=self.hidden_size,  # Hidden size for subsequent layers
            out_channels=self.hidden_size,  # Output channels
            kernel_size=3,
            padding="same",  # Padding to maintain sequence length
            bias=True,
            groups=self.hidden_size,
        )
        self.drop = nn.Dropout(dropout + 0.1)

        self.lskip = nn.Linear(self.hidden_size, self.hidden_size)

        self.wq = BlockDiagonal(
            in_features=self.hidden_size,
            out_features=self.hidden_size,
            num_blocks=num_layers,
            bias=bias,
        )
        self.wk = BlockDiagonal(
            in_features=self.hidden_size,
            out_features=self.hidden_size,
            num_blocks=num_layers,
            bias=bias,
        )
        self.wv = BlockDiagonal(
            in_features=self.hidden_size,
            out_features=self.hidden_size,
            num_blocks=num_layers,
            bias=bias,
        )
        self.dropq = nn.Dropout(dropout / 2)
        self.dropk = nn.Dropout(dropout / 2)
        self.dropv = nn.Dropout(dropout / 2)

        self.i_gate = nn.Linear(self.hidden_size, self.hidden_size)
        self.f_gate = nn.Linear(self.hidden_size, self.hidden_size)
        self.o_gate = nn.Linear(self.hidden_size, self.hidden_size)

        self.ln_c = nn.LayerNorm(self.hidden_size)
        self.ln_n = nn.LayerNorm(self.hidden_size)

        self.lnf = nn.LayerNorm(self.hidden_size)
        self.lno = nn.LayerNorm(self.hidden_size)
        self.lni = nn.LayerNorm(self.hidden_size)

        self.GN = nn.LayerNorm(self.hidden_size)
        self.ln_out = nn.LayerNorm(self.hidden_size)

        self.drop2 = nn.Dropout(dropout)

        self.proj = nn.Linear(self.hidden_size, self.hidden_size)
        self.ln_proj = nn.LayerNorm(self.hidden_size)

        # Remove fixed-size initializations for dynamic state initialization
        self.ct_1 = None
        self.nt_1 = None

    def init_states(self, batch_size, seq_length, device):
        """Initialize the state tensors with the correct batch and sequence dimensions.

        Parameters
        ----------
        batch_size : int
            The batch size.
        seq_length : int
            The sequence length.
        device : torch.device
            The device to place the tensors on.
        """
        self.ct_1 = torch.zeros(batch_size, seq_length, self.hidden_size, device=device)
        self.nt_1 = torch.zeros(batch_size, seq_length, self.hidden_size, device=device)

    def forward(self, x):
        """Forward pass through mLSTM block.

        Parameters
        ----------
        x : torch.Tensor
            Input tensor of shape (batch, sequence_length, input_size).

        Returns
        -------
        torch.Tensor
            Output tensor of shape (batch, sequence_length, input_size).
        """
        if x.ndim != 3:
            raise ValueError("Input tensor must have 3 dimensions (batch, sequence_length, input_size)")
        B, N, D = x.shape
        device = x.device

        # Initialize states dynamically based on input shape
        if self.ct_1 is None or self.ct_1.shape[0] != B or self.ct_1.shape[1] != N:
            self.init_states(B, N, device)

        x = self.ln(x)  # layer norm on x

        left = self.left(x)  # part left
        # part right with just swish (silu) function
        right = self.activation(self.right(x))

        left_left = left.transpose(1, 2)
        left_left = self.activation(self.drop(self.conv(left_left).transpose(1, 2)))
        l_skip = self.lskip(left_left)

        # start mLSTM
        q = self.dropq(self.wq(left_left))
        k = self.dropk(self.wk(left_left))
        v = self.dropv(self.wv(left))

        i = torch.exp(self.lni(self.i_gate(left_left)))
        f = torch.exp(self.lnf(self.f_gate(left_left)))
        o = torch.sigmoid(self.lno(self.o_gate(left_left)))

        ct_1 = self.ct_1

        ct = f * ct_1 + i * v * k
        ct = torch.mean(self.ln_c(ct), [0, 1], keepdim=True)
        self.ct_1 = ct.detach()

        nt_1 = self.nt_1
        nt = f * nt_1 + i * k
        nt = torch.mean(self.ln_n(nt), [0, 1], keepdim=True)
        self.nt_1 = nt.detach()

        ht = o * ((ct * q) / torch.max(nt * q))
        # end mLSTM
        ht = ht

        left = self.drop2(self.GN(ht + l_skip))

        out = self.ln_out(left * right)
        out = self.ln_proj(self.proj(out))

        return out, None


class sLSTMblock(nn.Module):
    """SLSTM block with convolutions, gated mechanisms, and projection layers.

    Parameters
    ----------
    input_size : int
        Size of the input features.
    hidden_size : int
        Size of the hidden state.
    num_layers : int
        Depth of block diagonal layers.
    dropout : float, optional
        Dropout probability (default is 0.2).
    """

    def __init__(
        self,
        input_size,
        hidden_size,
        num_layers,
        bidirectional=None,
        batch_first=None,
        nonlinearity=F.silu,
        dropout=0.2,
        bias=True,
    ):
        super().__init__()
        self.input_size = input_size
        self.hidden_size = hidden_size
        self.activation = nonlinearity

        self.drop = nn.Dropout(dropout)

        self.i_gate = BlockDiagonal(
            in_features=self.input_size,
            out_features=self.input_size,
            num_blocks=num_layers,
            bias=bias,
        )
        self.f_gate = BlockDiagonal(
            in_features=self.input_size,
            out_features=self.input_size,
            num_blocks=num_layers,
            bias=bias,
        )
        self.o_gate = BlockDiagonal(
            in_features=self.input_size,
            out_features=self.input_size,
            num_blocks=num_layers,
            bias=bias,
        )
        self.z_gate = BlockDiagonal(
            in_features=self.input_size,
            out_features=self.input_size,
            num_blocks=num_layers,
            bias=bias,
        )

        self.ri_gate = BlockDiagonal(self.input_size, self.input_size, num_layers, bias=False)
        self.rf_gate = BlockDiagonal(self.input_size, self.input_size, num_layers, bias=False)
        self.ro_gate = BlockDiagonal(self.input_size, self.input_size, num_layers, bias=False)
        self.rz_gate = BlockDiagonal(self.input_size, self.input_size, num_layers, bias=False)

        self.ln_i = nn.LayerNorm(self.input_size)
        self.ln_f = nn.LayerNorm(self.input_size)
        self.ln_o = nn.LayerNorm(self.input_size)
        self.ln_z = nn.LayerNorm(self.input_size)

        self.GN = nn.LayerNorm(self.input_size)
        self.ln_c = nn.LayerNorm(self.input_size)
        self.ln_n = nn.LayerNorm(self.input_size)
        self.ln_h = nn.LayerNorm(self.input_size)

        self.left_linear = nn.Linear(self.input_size, int(self.input_size * (4 / 3)))
        self.right_linear = nn.Linear(self.input_size, int(self.input_size * (4 / 3)))

        self.ln_out = nn.LayerNorm(int(self.input_size * (4 / 3)))

        self.proj = nn.Linear(int(self.input_size * (4 / 3)), self.hidden_size)

        # Remove initial fixed-size states
        self.ct_1 = None
        self.nt_1 = None
        self.ht_1 = None
        self.mt_1 = None

    def init_states(self, batch_size, seq_length, device):
        """Initialize the state tensors with the correct batch and sequence dimensions.

        Parameters
        ----------
        batch_size : int
            The batch size.
        seq_length : int
            The sequence length.
        device : torch.device
            The device to place the tensors on.
        """
        self.nt_1 = torch.zeros(batch_size, seq_length, self.input_size, device=device)
        self.ct_1 = torch.zeros(batch_size, seq_length, self.input_size, device=device)
        self.ht_1 = torch.zeros(batch_size, seq_length, self.input_size, device=device)
        self.mt_1 = torch.zeros(batch_size, seq_length, self.input_size, device=device)

    def forward(self, x):
        """Forward pass through sLSTM block.

        Parameters
        ----------
        x : torch.Tensor
            Input tensor of shape (batch, sequence_length, input_size).

        Returns
        -------
        torch.Tensor
            Output tensor of shape (batch, sequence_length, input_size).
        """
        B, N, D = x.shape
        device = x.device

        # Initialize states dynamically based on input shape
        if self.ct_1 is None or self.nt_1 is None or self.nt_1.shape[0] != B or self.nt_1.shape[1] != N:
            self.init_states(B, N, device)

        x = self.activation(x)

        # Start sLSTM operations
        ht_1 = self.ht_1

        i = torch.exp(self.ln_i(self.i_gate(x) + self.ri_gate(ht_1)))
        f = torch.exp(self.ln_f(self.f_gate(x) + self.rf_gate(ht_1)))

        # Use expand_as to match the shapes of f and i for element-wise operations
        m = torch.max(
            torch.log(f) + self.mt_1.expand_as(f),  # type: ignore
            torch.log(i),  # type: ignore
        )
        i = torch.exp(torch.log(i) - m)
        f = torch.exp(torch.log(f) + self.mt_1.expand_as(f) - m)  # type: ignore
        self.mt_1 = m.detach()

        o = torch.sigmoid(self.ln_o(self.o_gate(x) + self.ro_gate(ht_1)))
        z = torch.tanh(self.ln_z(self.z_gate(x) + self.rz_gate(ht_1)))

        ct_1 = self.ct_1
        ct = f * ct_1 + i * z
        ct = torch.mean(self.ln_c(ct), [0, 1], keepdim=True)
        self.ct_1 = ct.detach()

        nt_1 = self.nt_1
        nt = f * nt_1 + i
        nt = torch.mean(self.ln_n(nt), [0, 1], keepdim=True)
        self.nt_1 = nt.detach()

        ht = o * (ct / nt)
        ht = torch.mean(self.ln_h(ht), [0, 1], keepdim=True)
        self.ht_1 = ht.detach()

        slstm_out = self.GN(ht)

        left = self.left_linear(slstm_out)
        right = F.gelu(self.right_linear(slstm_out))

        out = self.ln_out(left * right)
        out = self.proj(out)
        return out, None



import torch
import torch.nn as nn


class Reshape(nn.Module):
    def __init__(self, j, dim, method="linear"):
        super().__init__()
        self.j = j
        self.dim = dim
        self.method = method

        if self.method == "linear":
            # Use nn.Linear approach
            self.layer = nn.Linear(dim, j * dim)
        elif self.method == "embedding":
            # Use nn.Embedding approach
            self.layer = nn.Embedding(dim, j * dim)
        elif self.method == "conv1d":
            # Use nn.Conv1d approach
            self.layer = nn.Conv1d(in_channels=dim, out_channels=j * dim, kernel_size=1)
        else:
            raise ValueError(f"Unsupported method '{method}' for reshaping.")

    def forward(self, x):
        batch_size = x.shape[0]

        if self.method == "linear" or self.method == "embedding":
            x_reshaped = self.layer(x)  # shape: (batch_size, j * dim)
            x_reshaped = x_reshaped.view(batch_size, self.j, self.dim)  # shape: (batch_size, j, dim)
        elif self.method == "conv1d":
            # For Conv1d, add dummy dimension and reshape
            x = x.unsqueeze(-1)  # Add dummy dimension for convolution
            x_reshaped = self.layer(x)  # shape: (batch_size, j * dim, 1)
            x_reshaped = x_reshaped.squeeze(-1)  # Remove dummy dimension
            x_reshaped = x_reshaped.view(batch_size, self.j, self.dim)  # shape: (batch_size, j, dim)

        return x_reshaped  # type: ignore


class AttentionNetBlock(nn.Module):
    def __init__(
        self,
        channels,
        in_channels,
        d_model,
        n_heads,
        n_layers,
        dim_feedforward,
        transformer_activation,
        output_dim,
        attn_dropout,
        layer_norm_eps,
        norm_first,
        bias,
        activation,
        embedding_activation,
        norm_f,
        method,
    ):
        super().__init__()

        self.reshape = Reshape(channels, in_channels, method)

        encoder_layer = nn.TransformerEncoderLayer(
            d_model=d_model,
            nhead=n_heads,
            batch_first=True,
            dim_feedforward=dim_feedforward,
            dropout=attn_dropout,
            activation=transformer_activation,
            layer_norm_eps=layer_norm_eps,
            norm_first=norm_first,
            bias=bias,
        )

        self.encoder = nn.TransformerEncoder(
            encoder_layer,
            num_layers=n_layers,
            norm=norm_f,
        )

        self.linear = nn.Linear(d_model, output_dim)
        self.activation = activation
        self.embedding_activation = embedding_activation

    def forward(self, x):
        z = self.reshape(x)
        x = self.embedding_activation(z)
        x = self.encoder(x)
        x = z + x
        x = torch.sum(x, dim=1)
        x = self.linear(x)
        x = self.activation(x)
        return x



def get_feature_dimensions(num_feature_info, cat_feature_info, embedding_info):
    input_dim = 0
    for _, feature_info in num_feature_info.items():
        input_dim += feature_info["dimension"]
    for _, feature_info in cat_feature_info.items():
        input_dim += feature_info["dimension"]
    for _, feature_info in embedding_info.items():
        input_dim += feature_info["dimension"]

    return input_dim



import math

import torch
import torch.nn as nn
import torch.nn.functional as F


class NeuralEmbeddingTree(nn.Module):
    def __init__(
        self,
        input_dim,
        output_dim,
        temperature=0.0,
    ):
        """Initialize the neural decision tree with a neural network at each leaf.

        Parameters:
        -----------
        input_dim: int
            The number of input features.
        depth: int
            The depth of the tree. The number of leaves will be 2^depth.
        output_dim: int
            The number of output classes (default is 1 for regression tasks).
        lamda: float
            Regularization parameter.
        """
        super().__init__()

        self.temperature = temperature
        self.output_dim = output_dim
        self.depth = int(math.log2(output_dim))

        # Initialize internal nodes with linear layers followed by hard thresholds
        self.inner_nodes = nn.Sequential(
            nn.Linear(input_dim + 1, output_dim, bias=False),
        )

    def forward(self, X):
        """Implementation of the forward pass with hard decision boundaries."""
        batch_size = X.size()[0]
        X = self._data_augment(X)

        # Get the decision boundaries for the internal nodes
        decision_boundaries = self.inner_nodes(X)

        # Apply hard thresholding to simulate binary decisions
        if self.temperature > 0.0:
            # Replace sigmoid with Gumbel-Softmax for path_prob calculation
            logits = decision_boundaries / self.temperature
            path_prob = (logits > 0).float() + logits.sigmoid() - logits.sigmoid().detach()
        else:
            path_prob = (decision_boundaries > 0).float()

        # Prepare for routing at the internal nodes
        path_prob = torch.unsqueeze(path_prob, dim=2)
        path_prob = torch.cat((path_prob, 1 - path_prob), dim=2)

        _mu = X.data.new(batch_size, 1, 1).fill_(1.0)

        # Iterate through internal nodes in each layer to compute the final path
        # probabilities and the regularization term.
        begin_idx = 0
        end_idx = 1

        for layer_idx in range(0, self.depth):
            _path_prob = path_prob[:, begin_idx:end_idx, :]

            _mu = _mu.view(batch_size, -1, 1).repeat(1, 1, 2)

            _mu = _mu * _path_prob  # update path probabilities

            begin_idx = end_idx
            end_idx = begin_idx + 2 ** (layer_idx + 1)

        mu = _mu.view(batch_size, self.output_dim)

        return mu

    def _data_augment(self, X):
        return F.pad(X, (1, 0), value=1)



import torch.nn as nn
from ..arch_utils.layer_utils.embedding_layer import EmbeddingLayer
from .utils.basemodel import BaseModel
import torch.nn.init as nn_init
import numpy as np
from ..configs.autoint_config import DefaultAutoIntConfig


class AutoInt(BaseModel):
    """
    AutoInt: Automatic Feature Interaction Learning via Self-Attentive Neural Networks.

    This model uses multi-head self-attention layers to learn feature interactions for tabular data.
    It supports key-value compression for memory efficiency and is compatible with embedding-based
    feature encodings.

    Parameters
    ----------
    feature_information : tuple
        A tuple containing information about numerical features, categorical features,
        and any additional embeddings. Expected format: `(num_feature_info, cat_feature_info, embedding_feature_info)`.
    num_classes : int, default=1
        Number of output classes. For regression, this should be set to `1`.
    config : DefaultAutoIntConfig, optional
        Configuration object containing hyperparameters such as `d_model`, `n_heads`, `n_layers`,
        dropout rates, and compression settings.
    **kwargs : dict
        Additional arguments passed to the `BaseModel`.

    Attributes
    ----------
    embedding_layer : EmbeddingLayer
        Module that processes numerical and categorical features into embeddings.
    kv_compression : float or None
        The proportion of key-value compression. If `None`, no compression is applied.
    kv_compression_sharing : str or None
        Defines how key-value compression is shared across layers. Options:
        - `"layerwise"`: One shared compression layer for all layers.
        - `"headwise"`: Separate key compression per head.
        - `"key-value"`: Separate compression layers for `k` and `v`.
    shared_kv_compression : nn.Linear or None
        Shared key-value compression layer, used when `kv_compression_sharing="layerwise"`.
    layers : nn.ModuleList
        A list of transformer-based attention layers, each consisting of:
        - `attention`: Multi-head self-attention module.
        - `linear`: Fully connected layer for projection.
        - `norm0`: Layer normalization.
    last_norm : nn.LayerNorm or None
        Final normalization layer applied before output if `prenormalization` is enabled.
    head : nn.Linear
        Output layer mapping from the processed feature representation to the final predictions.
    """

    def __init__(
        self,
        feature_information: tuple,  # (num_feature_info, cat_feature_info, embedding_feature_info)
        num_classes=1,
        config: DefaultAutoIntConfig = DefaultAutoIntConfig(),  # noqa: B008
        **kwargs,
    ):
        super().__init__(config=config, **kwargs)
        self.save_hyperparameters(ignore=["feature_information"])
        self.returns_ensemble = False

        # Embedding layer
        self.embedding_layer = EmbeddingLayer(*feature_information, config=config)
        n_inputs = np.sum([len(info) for info in feature_information])

        # Key-Value Compression
        self.kv_compression = config.kv_compression
        self.kv_compression_sharing = config.kv_compression_sharing

        def make_kv_compression():
            compression = nn.Linear(
                n_inputs,
                int(n_inputs * config.kv_compression),
                bias=False,
            )
            nn_init.xavier_uniform_(compression.weight)
            return compression

        self.shared_kv_compression = (
            make_kv_compression()
            if self.kv_compression and self.kv_compression_sharing == "layerwise"
            else None
        )

        # Transformer-based Interaction Layers
        self.layers = nn.ModuleList()
        for layer_idx in range(config.n_layers):
            layer = nn.ModuleDict(
                {
                    "attention": nn.MultiheadAttention(
                        embed_dim=config.d_model,
                        num_heads=config.n_heads,
                        dropout=config.attn_dropout,
                        batch_first=True,
                    ),
                    "linear": nn.Linear(config.d_model, config.d_model, bias=False),
                    "norm0": nn.LayerNorm(config.d_model),
                }
            )

            if self.kv_compression and self.shared_kv_compression is None:
                layer["key_compression"] = make_kv_compression()
                if self.kv_compression_sharing == "headwise":
                    layer["value_compression"] = make_kv_compression()
                else:
                    assert self.kv_compression_sharing == "key-value"

            self.layers.append(layer)

        # Final Normalization & Output Head
        self.last_norm = (
            nn.LayerNorm(config.d_model) if getattr(config, "prenorm", False) else None
        )

        self.head = nn.Linear(config.d_model * n_inputs, num_classes)

    def _get_kv_compressions(self, layer):
        """
        Returns the correct key-value compression layers based on the sharing strategy.

        Parameters
        ----------
        layer : nn.ModuleDict
            The transformer layer containing possible key-value compression modules.

        Returns
        -------
        tuple of (nn.Linear or None, nn.Linear or None)
            The key compression and value compression layers, or `(None, None)` if no compression is applied.
        """
        return (
            (self.shared_kv_compression, self.shared_kv_compression)
            if self.shared_kv_compression is not None
            else (
                (layer["key_compression"], layer["value_compression"])
                if "key_compression" in layer and "value_compression" in layer
                else (
                    (layer["key_compression"], layer["key_compression"])
                    if "key_compression" in layer
                    else (None, None)
                )
            )
        )

    def forward(self, *data):
        """
        Forward pass of the AutoInt model.

        Parameters
        ----------
        *data : tuple
            Input tuple of tensors containing numerical features, categorical features, and embeddings.

        Returns
        -------
        Tensor
            The output predictions of the model.
        """
        x = self.embedding_layer(*data)  # Shape: (N, J, d_model)

        for layer in self.layers:
            x_residual = x  # Store original input for residual connection

            # Apply normalization before attention if prenormalization is enabled
            x_residual = layer["norm0"](x_residual)

            # Retrieve key-value compression layers
            key_compression, value_compression = self._get_kv_compressions(layer)

            # Multihead Attention
            x_residual, _ = layer["attention"](x_residual, x_residual, x_residual)

            # Apply residual connection
            x = x + x_residual

            # Apply the linear transformation
            x_residual = layer["linear"](x)
            x = x + x_residual  # Second residual connection

        if self.last_norm:
            x = self.last_norm(x)  # Final normalization if prenormalization is used

        x = x.flatten(1)  # Flatten from (N, J, d_model) to (N, J * d_model)
        return self.head(x)  # Final prediction



# ruff: noqa

import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
from einops import rearrange


class GEGLU(nn.Module):
    def forward(self, x):
        x, gates = x.chunk(2, dim=-1)
        return x * F.gelu(gates)


def FeedForward(dim, mult=4, dropout=0.0):
    return nn.Sequential(
        nn.LayerNorm(dim),
        nn.Linear(dim, dim * mult * 2),
        GEGLU(),
        nn.Dropout(dropout),
        nn.Linear(dim * mult, dim),
    )


class Attention(nn.Module):
    def __init__(self, dim, heads=8, dim_head=64, dropout=0.0):
        super().__init__()
        inner_dim = dim_head * heads
        self.heads = heads
        self.scale = dim_head**-0.5
        self.norm = nn.LayerNorm(dim)
        self.to_qkv = nn.Linear(dim, inner_dim * 3, bias=False)
        self.to_out = nn.Linear(inner_dim, dim, bias=False)
        self.dropout = nn.Dropout(dropout)
        dim = np.int64(dim / 2)

    def forward(self, x):
        h = self.heads
        x = self.norm(x)
        q, k, v = self.to_qkv(x).chunk(3, dim=-1)
        q, k, v = map(lambda t: rearrange(t, "b n (h d) -> b h n d", h=h), (q, k, v))  # type: ignore
        q = q * self.scale

        sim = torch.einsum("b h i d, b h j d -> b h i j", q, k)

        attn = sim.softmax(dim=-1)
        dropped_attn = self.dropout(attn)

        out = torch.einsum("b h i j, b h j d -> b h i d", dropped_attn, v)
        out = rearrange(out, "b h n d -> b n (h d)", h=h)
        out = self.to_out(out)

        return out, attn


class Transformer(nn.Module):
    def __init__(self, dim, depth, heads, dim_head, attn_dropout, ff_dropout):
        super().__init__()
        self.layers = nn.ModuleList([])

        for _ in range(depth):
            self.layers.append(
                nn.ModuleList(
                    [
                        Attention(
                            dim,
                            heads=heads,
                            dim_head=dim_head,
                            dropout=attn_dropout,
                        ),
                        FeedForward(dim, dropout=ff_dropout),
                    ]
                )
            )

    def forward(self, x, return_attn=False):
        post_softmax_attns = []

        for attn, ff in self.layers:  # type: ignore
            attn_out, post_softmax_attn = attn(x)
            post_softmax_attns.append(post_softmax_attn)

            x = attn_out + x
            x = ff(x) + x

        if not return_attn:
            return x

        return x, torch.stack(post_softmax_attns)



import math

import torch
import torch.nn as nn
from torch.nn.parameter import Parameter

from .sn_linear import SNLinear


class Periodic(nn.Module):
    """Periodic transformation with learned frequency coefficients."""

    def __init__(self, n_features: int, k: int, sigma: float) -> None:
        super().__init__()
        if sigma <= 0.0:
            raise ValueError(f"sigma must be positive, but got {sigma=}")

        self._sigma = sigma
        self.weight = Parameter(torch.empty(n_features, k))
        self.reset_parameters()

    def reset_parameters(self) -> None:
        bound = self._sigma * 3
        nn.init.trunc_normal_(self.weight, 0.0, self._sigma, a=-bound, b=bound)

    def forward(self, x):
        x = 2 * math.pi * self.weight * x[..., None]
        return torch.cat([torch.cos(x), torch.sin(x)], dim=-1)


class PeriodicEmbeddings(nn.Module):
    """Embeddings for continuous features using Periodic + Linear (+ ReLU) transformations.

    Supports PL, PLR, and PLR(lite) embedding types.

    Shape:
        - Input: (*, n_features)
        - Output: (*, n_features, d_embedding)
    """

    def __init__(
        self,
        n_features: int,
        d_embedding: int = 24,
        *,
        n_frequencies: int = 48,
        frequency_init_scale: float = 0.01,
        activation: bool = True,
        lite: bool = False,
    ):
        """
        Args:
            n_features (int): Number of features.
            d_embedding (int): Size of each feature embedding.
            n_frequencies (int): Number of frequencies per feature.
            frequency_init_scale (float): Initialization scale for frequency coefficients.
            activation (bool): If True, applies ReLU, making it PLR; otherwise, PL.
            lite (bool): If True, uses shared linear layer (PLR lite); otherwise, separate layers.
        """
        super().__init__()
        self.periodic = Periodic(n_features, n_frequencies, frequency_init_scale)

        # Choose linear transformation: shared or separate
        if lite:
            if not activation:
                raise ValueError("lite=True requires activation=True")
            self.linear = nn.Linear(2 * n_frequencies, d_embedding)
        else:
            self.linear = SNLinear(n_features, 2 * n_frequencies, d_embedding)

        self.activation = nn.ReLU() if activation else None

    def forward(self, x):
        """Forward pass."""
        x = self.periodic(x)
        x = self.linear(x)
        return self.activation(x) if self.activation else x



from ..base_models.enode import ENODE
from ..configs.enode_config import DefaultENODEConfig
from ..utils.docstring_generator import generate_docstring
from .utils.sklearn_base_classifier import SklearnBaseClassifier
from .utils.sklearn_base_lss import SklearnBaseLSS
from .utils.sklearn_base_regressor import SklearnBaseRegressor


class ENODERegressor(SklearnBaseRegressor):
    __doc__ = generate_docstring(
        DefaultENODEConfig,
        model_description="""
        Neural Oblivious Decision Ensemble (ENODE) Regressor. Slightly different with a MLP as a tabular task specific head. This class extends the SklearnBaseRegressor class and uses the ENODE model
        with the default ENODE configuration.
        """,
        examples="""
        >>> from mambular.models import ENODERegressor
        >>> model = ENODERegressor()
        >>> model.fit(X_train, y_train)
        >>> preds = model.predict(X_test)
        >>> model.evaluate(X_test, y_test)
        """,
    )

    def __init__(self, **kwargs):
        super().__init__(model=ENODE, config=DefaultENODEConfig, **kwargs)


class ENODEClassifier(SklearnBaseClassifier):
    __doc__ = generate_docstring(
        DefaultENODEConfig,
        model_description="""
        Neural Oblivious Decision Ensemble (ENODE) Classifier. Slightly different with a MLP as a tabular task specific head.
        This class extends the SklearnBaseClassifier class and uses the ENODE model
        with the default ENODE configuration.
        """,
        examples="""
        >>> from mambular.models import ENODEClassifier
        >>> model = ENODEClassifier()
        >>> model.fit(X_train, y_train)
        >>> preds = model.predict(X_test)
        >>> model.evaluate(X_test, y_test)
        """,
    )

    def __init__(self, **kwargs):
        super().__init__(model=ENODE, config=DefaultENODEConfig, **kwargs)


class ENODELSS(SklearnBaseLSS):
    __doc__ = generate_docstring(
        DefaultENODEConfig,
        model_description="""
        Neural Oblivious Decision Ensemble (ENODE) for distributional regression. Slightly different with a MLP as a tabular task specific head.
        This class extends the SklearnBaseLSS class and uses the ENODE model
        with the default ENODE configuration.
        """,
        examples="""
        >>> from mambular.models import ENODELSS
        >>> model = ENODELSS()
        >>> model.fit(X_train, y_train, family='normal')
        >>> preds = model.predict(X_test)
        >>> model.evaluate(X_test, y_test)
        """,
    )

    def __init__(self, **kwargs):
        super().__init__(model=ENODE, config=DefaultENODEConfig, **kwargs)



from collections.abc import Callable
from dataclasses import dataclass, field
from typing import Literal
import torch.nn as nn
from .base_config import BaseConfig


@dataclass
class DefaultTabMConfig(BaseConfig):
    """Configuration class for the TabM model with batch ensembling and predefined hyperparameters.

    Parameters
    ----------
    layer_sizes : list, default=(512, 512, 128)
        Sizes of the layers in the model.
    activation : callable, default=nn.ReLU()
        Activation function for the model layers.
    dropout : float, default=0.3
        Dropout rate for regularization.
    norm : str, default=None
        Normalization method to be used, if any.
    use_glu : bool, default=False
        Whether to use Gated Linear Units (GLU) in the model.
    ensemble_size : int, default=32
        Number of ensemble members for batch ensembling.
    ensemble_scaling_in : bool, default=True
        Whether to use input scaling for each ensemble member.
    ensemble_scaling_out : bool, default=True
        Whether to use output scaling for each ensemble member.
    ensemble_bias : bool, default=True
        Whether to use a unique bias term for each ensemble member.
    scaling_init : {"ones", "random-signs", "normal"}, default="normal"
        Initialization method for scaling weights.
    average_ensembles : bool, default=False
        Whether to average the outputs of the ensembles.
    model_type : {"mini", "full"}, default="mini"
        Model type to use ('mini' for reduced version, 'full' for complete model).
    """

    # arch params
    layer_sizes: list = field(default_factory=lambda: [256, 256, 128])
    activation: Callable = nn.ReLU()  # noqa: RUF009
    dropout: float = 0.5
    norm: str | None = None
    use_glu: bool = False

    # Batch ensembling specific configurations
    ensemble_size: int = 32
    ensemble_scaling_in: bool = True
    ensemble_scaling_out: bool = True
    ensemble_bias: bool = True
    scaling_init: Literal["ones", "random-signs", "normal"] = "ones"
    average_ensembles: bool = False
    model_type: Literal["mini", "full"] = "mini"
    average_embeddings: bool = True



import torch
import torch.nn as nn
import torch.nn.functional as F
from mambular.arch_utils.layer_utils.sparsemax import sparsemax, sparsemoid
from .data_aware_initialization import ModuleWithInit
from .numpy_utils import check_numpy
import numpy as np
from warnings import warn


class ODSTE(ModuleWithInit):

    def __init__(
        self,
        in_features,  # J (number of features)
        num_trees,
        embed_dim,  # D (embedding dimension per feature)
        depth=6,
        tree_dim=1,
        flatten_output=True,
        choice_function=sparsemax,
        bin_function=sparsemoid,
        initialize_response_=nn.init.normal_,
        initialize_selection_logits_=nn.init.uniform_,
        threshold_init_beta=1.0,
        threshold_init_cutoff=1.0,
    ):
        """Oblivious Differentiable Sparsemax Trees (ODST) with Feature & Embedding Splitting."""
        super().__init__()
        self.depth, self.num_trees, self.tree_dim, self.flatten_output = (
            depth,
            num_trees,
            tree_dim,
            flatten_output,
        )
        self.choice_function, self.bin_function = choice_function, bin_function
        self.in_features, self.embed_dim = in_features, embed_dim
        self.threshold_init_beta, self.threshold_init_cutoff = (
            threshold_init_beta,
            threshold_init_cutoff,
        )

        # Response values for each leaf
        self.response = nn.Parameter(
            torch.zeros([num_trees, tree_dim, embed_dim, 2**depth]), requires_grad=True
        )

        initialize_response_(self.response)

        # Feature selection logits (choose J)
        self.feature_selection_logits = nn.Parameter(
            torch.zeros([num_trees, depth, in_features]), requires_grad=True
        )
        initialize_selection_logits_(self.feature_selection_logits)

        # Embedding selection logits (choose D within J)
        self.embedding_selection_logits = nn.Parameter(
            torch.randn([num_trees, depth, in_features, embed_dim])
        )

        # Thresholds & temperatures (random initialization)
        self.feature_thresholds = nn.Parameter(torch.randn([num_trees, depth]))
        self.log_temperatures = nn.Parameter(torch.randn([num_trees, depth]))

        # Binary code mappings
        with torch.no_grad():
            indices = torch.arange(2**self.depth)
            offsets = 2 ** torch.arange(self.depth)
            bin_codes = (indices.view(1, -1) // offsets.view(-1, 1) % 2).to(
                torch.float32
            )
            bin_codes_1hot = torch.stack([bin_codes, 1.0 - bin_codes], dim=-1)
            self.bin_codes_1hot = nn.Parameter(bin_codes_1hot, requires_grad=False)

    def initialize(self, x, eps=1e-6):
        """Data-aware initialization of thresholds and log-temperatures based on input data.

        Parameters
        ----------
        x : torch.Tensor
            Input tensor of shape [batch_size, in_features, embed_dim] used for threshold initialization.
        eps : float, optional
            Small value added to avoid log(0) errors in temperature initialization. Default is 1e-6.
        """
        if len(x.shape) != 3:
            raise ValueError("Input tensor must have shape (batch_size, J, D)")

        if x.shape[0] < 1000:
            warn(
                "Data-aware initialization is performed on less than 1000 data points. This may cause instability."
                "To avoid potential problems, run this model on a data batch with at least 1000 data samples."
                "You can do so manually before training. Use with torch.no_grad() for memory efficiency."
            )

        with torch.no_grad():
            # Select features (J)
            feature_selectors = self.choice_function(
                self.feature_selection_logits, dim=-1
            )
            # feature_selectors shape: (num_trees, depth, J)

            selected_features = torch.einsum("bjd,ntj->bntd", x, feature_selectors)
            # selected_features shape: (B, num_trees, depth, D)

            # Select embeddings (D)
            embedding_selectors = self.choice_function(
                self.embedding_selection_logits, dim=-1
            )
            # embedding_selectors shape: (num_trees, depth, J, D)

            selected_embeddings = torch.einsum(
                "bntd,ntjd->bntd", selected_features, embedding_selectors
            )
            # selected_embeddings shape: (B, num_trees, depth, D)

            # Initialize thresholds using percentiles from the data
            percentiles_q = 100 * np.random.beta(
                self.threshold_init_beta,
                self.threshold_init_beta,
                size=[self.num_trees, self.depth],
            )

            reshaped_embeddings = selected_embeddings.permute(1, 2, 0, 3).reshape(
                self.num_trees * self.depth, -1
            )
            self.feature_thresholds.data[...] = torch.as_tensor(
                list(
                    map(
                        np.percentile,
                        check_numpy(reshaped_embeddings),  # Now correctly 2D
                        percentiles_q.flatten(),
                    )
                ),
                dtype=selected_embeddings.dtype,
                device=selected_embeddings.device,
            ).view(self.num_trees, self.depth)

            # Initialize temperatures based on the threshold differences
            temperatures = np.percentile(
                check_numpy(
                    abs(selected_embeddings - self.feature_thresholds.unsqueeze(-1))
                ),
                q=100 * min(1.0, self.threshold_init_cutoff),
                axis=0,
            )

            # Scale temperatures based on the cutoff
            temperatures /= max(1.0, self.threshold_init_cutoff)

            self.log_temperatures.data[...] = torch.log(
                torch.as_tensor(
                    temperatures.mean(-1),
                    dtype=selected_embeddings.dtype,
                    device=selected_embeddings.device,
                )
                + eps
            )

    def forward(self, x):
        if len(x.shape) != 3:
            raise ValueError("Input tensor must have shape (batch_size, J, D)")

        # Select feature (J) and embedding dimension (D) separately
        feature_selectors = self.choice_function(
            self.feature_selection_logits, dim=-1
        )  # [num_trees, depth, J]

        embedding_selectors = self.choice_function(
            self.embedding_selection_logits, dim=-1
        )  # [num_trees, depth, J, D]

        # Select features (J) first
        selected_features = torch.einsum("bjd,ntj->bntd", x, feature_selectors)

        # Select embeddings (D) within selected features
        selected_embeddings = torch.einsum(
            "bntd,ntjd->bntd", selected_features, embedding_selectors
        )

        # Compute threshold logits
        threshold_logits = (
            selected_embeddings - self.feature_thresholds.unsqueeze(0).unsqueeze(-1)
        ) * torch.exp(-self.log_temperatures.unsqueeze(0).unsqueeze(-1))

        threshold_logits = torch.stack([-threshold_logits, threshold_logits], dim=-1)

        # Compute binary decisions
        bins = self.bin_function(threshold_logits)

        bin_matches = torch.einsum("bntds,tcs->bntdc", bins, self.bin_codes_1hot)

        response_weights = torch.prod(bin_matches, dim=2)

        # Compute final response
        response = torch.einsum("bnds,ncds->bnd", response_weights, self.response)
        return response

    def __repr__(self):
        return f"{self.__class__.__name__}(in_features={self.in_features}, embed_dim={self.embed_dim}, num_trees={self.num_trees}, depth={self.depth}, tree_dim={self.tree_dim}, flatten_output={self.flatten_output})"


class DenseBlock(nn.Module):
    """DenseBlock that sequentially stacks attention layers and `Module` layers (e.g., ODSTE)
    with feature and embedding-aware splits.

    Parameters
    ----------
    input_dim : int
        Number of features (J) in the input.
    embed_dim : int
        Embedding dimension per feature (D).
    layer_dim : int
        Dimensionality of each ODSTE layer.
    num_layers : int
        Number of layers to stack in the block.
    tree_dim : int, optional
        Number of output channels from each tree. Default is 1.
    max_features : int, optional
        Maximum number of features for expansion. Default is None.
    input_dropout : float, optional
        Dropout rate applied to inputs during training. Default is 0.0.
    flatten_output : bool, optional
        If True, flattens the output along the tree dimension. Default is True.
    Module : nn.Module, optional
        Module class to use for each layer in the block. Default is `ODSTE`.
    **kwargs : dict
        Additional keyword arguments for `Module` instances.
    """

    def __init__(
        self,
        input_dim,
        embed_dim,
        layer_dim,
        num_layers,
        tree_dim=1,
        max_features=None,
        input_dropout=0.0,
        flatten_output=True,
        Module=ODSTE,
        **kwargs,
    ):
        super().__init__()
        self.num_layers = num_layers
        self.layer_dim = layer_dim
        self.tree_dim = tree_dim
        self.max_features = max_features
        self.input_dropout = input_dropout
        self.flatten_output = flatten_output

        self.attention_layers = nn.ModuleList()
        self.odste_layers = nn.ModuleList()

        for _ in range(num_layers):
            # self.attention_layers.append(
            #    nn.MultiheadAttention(
            #        embed_dim=embed_dim, num_heads=1, batch_first=True
            #    )
            # )
            self.odste_layers.append(
                Module(
                    in_features=input_dim,
                    embed_dim=embed_dim,
                    num_trees=layer_dim,
                    tree_dim=tree_dim,
                    flatten_output=True,
                    **kwargs,
                )
            )
            input_dim = min(
                input_dim + layer_dim * tree_dim, max_features or float("inf")
            )

    def forward(self, x):
        """Forward pass through the DenseBlock.

        Parameters
        ----------
        x : torch.Tensor
            Input tensor of shape [batch_size, J, D].

        Returns
        -------
        torch.Tensor
            Output tensor with expanded features.
        """
        initial_features = x.shape[1]  # J (num features)

        for odste_layer in self.odste_layers:
            # x, _ = attn_layer(x, x, x)  # Apply attention

            if self.max_features is not None:
                tail_features = min(self.max_features, x.shape[1]) - initial_features
                if tail_features > 0:
                    x = torch.cat(
                        [x[:, :initial_features, :], x[:, -tail_features:, :]], dim=1
                    )

            if self.training and self.input_dropout:
                x = F.dropout(x, self.input_dropout)

            h = odste_layer(x)  # Apply ODSTE layer
            x = torch.cat([x, h], dim=1)  # Concatenate new features

        return x



import torch
import torch.nn as nn


class BlockDiagonal(nn.Module):
    def __init__(self, in_features, out_features, num_blocks, bias=True):
        super().__init__()
        self.in_features = in_features
        self.out_features = out_features
        self.num_blocks = num_blocks

        if out_features % num_blocks != 0:
            raise ValueError("out_features must be divisible by num_blocks")

        block_out_features = out_features // num_blocks

        self.blocks = nn.ModuleList([nn.Linear(in_features, block_out_features, bias=bias) for _ in range(num_blocks)])

    def forward(self, x):
        x = [block(x) for block in self.blocks]
        x = torch.cat(x, dim=-1)
        return x



from dataclasses import dataclass, field
from collections.abc import Callable
import torch.nn as nn


@dataclass
class BaseConfig:
    """
    Base configuration class with shared hyperparameters for models.

    This configuration class provides common hyperparameters for optimization,
    embeddings, and categorical encoding, which can be inherited by specific
    model configurations.

    Parameters
    ----------
    lr : float, default=1e-04
        Learning rate for the optimizer.
    lr_patience : int, default=10
        Number of epochs with no improvement before reducing the learning rate.
    weight_decay : float, default=1e-06
        L2 regularization parameter for weight decay in the optimizer.
    lr_factor : float, default=0.1
        Factor by which the learning rate is reduced when patience is exceeded.
    activation : Callable, default=nn.ReLU()
        Activation function to use in the model's layers.
    cat_encoding : str, default="int"
        Method for encoding categorical features ('int', 'one-hot', or 'linear').

    Embedding Parameters
    --------------------
    use_embeddings : bool, default=False
        Whether to use embeddings for categorical or numerical features.
    embedding_activation : Callable, default=nn.Identity()
        Activation function applied to embeddings.
    embedding_type : str, default="linear"
        Type of embedding to use ('linear', 'plr', etc.).
    embedding_bias : bool, default=False
        Whether to use bias in embedding layers.
    layer_norm_after_embedding : bool, default=False
        Whether to apply layer normalization after embedding layers.
    d_model : int, default=32
        Dimensionality of embeddings or model representations.
    plr_lite : bool, default=False
        Whether to use a lightweight version of Piecewise Linear Regression (PLR).
    n_frequencies : int, default=48
        Number of frequency components for embeddings.
    frequencies_init_scale : float, default=0.01
        Initial scale for frequency components in embeddings.
    embedding_projection : bool, default=True
        Whether to apply a projection layer after embeddings.

    Notes
    -----
    - This base class is meant to be inherited by other configurations.
    - Provides default values that can be overridden in derived configurations.

    """

    # Training Parameters
    lr: float = 1e-04
    lr_patience: int = 10
    weight_decay: float = 1e-06
    lr_factor: float = 0.1

    # Embedding Parameters
    use_embeddings: bool = False
    embedding_activation: Callable = nn.Identity()  # noqa: RUF009
    embedding_type: str = "linear"
    embedding_bias: bool = False
    layer_norm_after_embedding: bool = False
    d_model: int = 32
    plr_lite: bool = False
    n_frequencies: int = 48
    frequencies_init_scale: float = 0.01
    embedding_projection: bool = True

    # Architecture Parameters
    batch_norm: bool = False
    layer_norm: bool = False
    layer_norm_eps: float = 1e-05
    activation: Callable = nn.ReLU()  # noqa: RUF009
    cat_encoding: str = "int"



from ..base_models.modern_nca import ModernNCA
from ..configs.modernnca_config import DefaultModernNCAConfig
from ..utils.docstring_generator import generate_docstring
from .utils.sklearn_base_classifier import SklearnBaseClassifier
from .utils.sklearn_base_lss import SklearnBaseLSS
from .utils.sklearn_base_regressor import SklearnBaseRegressor


class ModernNCARegressor(SklearnBaseRegressor):
    __doc__ = generate_docstring(
        DefaultModernNCAConfig,
        model_description="""
        Multi-Layer Perceptron regressor. This class extends the SklearnBaseRegressor class and uses the ModernNCA model
        with the default ModernNCA configuration.
        """,
        examples="""
        >>> from mambular.models import ModernNCARegressor
        >>> model = ModernNCARegressor(d_model=64, n_layers=8)
        >>> model.fit(X_train, y_train)
        >>> preds = model.predict(X_test)
        >>> model.evaluate(X_test, y_test)
        """,
    )

    def __init__(self, **kwargs):
        super().__init__(model=ModernNCA, config=DefaultModernNCAConfig, **kwargs)


class ModernNCAClassifier(SklearnBaseClassifier):
    __doc__ = generate_docstring(
        DefaultModernNCAConfig,
        model_description="""
        Multi-Layer Perceptron classifier This class extends the SklearnBaseClassifier class and uses the ModernNCA model
        with the default ModernNCA configuration.
        """,
        examples="""
        >>> from mambular.models import ModernNCAClassifier
        >>> model = ModernNCAClassifier(d_model=64, n_layers=8)
        >>> model.fit(X_train, y_train)
        >>> preds = model.predict(X_test)
        >>> model.evaluate(X_test, y_test)
        """,
    )

    def __init__(self, **kwargs):
        super().__init__(model=ModernNCA, config=DefaultModernNCAConfig, **kwargs)


class ModernNCALSS(SklearnBaseLSS):
    __doc__ = generate_docstring(
        DefaultModernNCAConfig,
        model_description="""
        Multi-Layer Perceptron for distributional regression. This class extends the SklearnBaseLSS class and uses the ModernNCA model
        with the default ModernNCA configuration.
        """,
        examples="""
        >>> from mambular.models import ModernNCALSS
        >>> model = ModernNCALSS(d_model=64, n_layers=8)
        >>> model.fit(X_train, y_train, family='normal')
        >>> preds = model.predict(X_test)
        >>> model.evaluate(X_test, y_test)
        """,
    )

    def __init__(self, **kwargs):
        super().__init__(model=ModernNCA, config=DefaultModernNCAConfig, **kwargs)



from dataclasses import replace
import torch
import torch.nn as nn

from ..arch_utils.get_norm_fn import get_normalization_layer
from ..arch_utils.layer_utils.embedding_layer import EmbeddingLayer
from ..arch_utils.mlp_utils import MLPhead
from ..arch_utils.rnn_utils import ConvRNN
from ..configs.tabularnn_config import DefaultTabulaRNNConfig
from .utils.basemodel import BaseModel


class TabulaRNN(BaseModel):

    def __init__(
        self,
        feature_information: tuple,  # Expecting (num_feature_info, cat_feature_info, embedding_feature_info)
        num_classes=1,
        config: DefaultTabulaRNNConfig = DefaultTabulaRNNConfig(),  # noqa: B008
        **kwargs,
    ):
        super().__init__(config=config, **kwargs)
        self.save_hyperparameters(ignore=["feature_information"])

        self.returns_ensemble = False

        self.rnn = ConvRNN(config)

        self.embedding_layer = EmbeddingLayer(
            *feature_information,
            config=config,
        )

        self.tabular_head = MLPhead(
            input_dim=self.hparams.dim_feedforward,
            config=config,
            output_dim=num_classes,
        )

        self.linear = nn.Linear(
            self.hparams.d_model,
            self.hparams.dim_feedforward,
        )

        temp_config = replace(config, d_model=config.dim_feedforward)
        self.norm_f = get_normalization_layer(temp_config)

        # pooling
        n_inputs = [len(info) for info in feature_information]
        self.initialize_pooling_layers(config=config, n_inputs=n_inputs)

    def forward(self, *data):
        """Defines the forward pass of the model.

        Parameters
        ----------
        num_features : Tensor
            Tensor containing the numerical features.
        cat_features : Tensor
            Tensor containing the categorical features.

        Returns
        -------
        Tensor
            The output predictions of the model.
        """

        x = self.embedding_layer(*data)
        # RNN forward pass
        out, _ = self.rnn(x)
        z = self.linear(torch.mean(x, dim=1))

        x = self.pool_sequence(out)
        x = x + z
        if self.norm_f is not None:
            x = self.norm_f(x)
        preds = self.tabular_head(x)

        return preds



import torch.nn as nn

from ..get_norm_fn import get_normalization_layer
from .mamba_arch import ResidualBlock


class MambAttn(nn.Module):
    """Mamba model composed of alternating MambaBlocks and Attention layers.

    Attributes:
        config (MambaConfig): Configuration object for the Mamba model.
        layers (nn.ModuleList): List of alternating ResidualBlock (Mamba layers) and
        attention layers constituting the model.
    """

    def __init__(
        self,
        config,
    ):
        super().__init__()

        # Define Mamba and Attention layers alternation
        self.layers = nn.ModuleList()

        total_blocks = config.n_layers + config.n_attention_layers  # Total blocks to be created
        attention_count = 0

        for i in range(total_blocks):
            # Insert attention layer after N Mamba layers
            if (i + 1) % (config.n_mamba_per_attention + 1) == 0:
                self.layers.append(
                    nn.MultiheadAttention(
                        embed_dim=config.d_model,
                        num_heads=config.n_heads,
                        dropout=config.attn_dropout,
                    )
                )
                attention_count += 1
            else:
                self.layers.append(
                    ResidualBlock(
                        d_model=config.d_model,
                        expand_factor=config.expand_factor,
                        bias=config.bias,
                        d_conv=config.d_conv,
                        conv_bias=config.conv_bias,
                        dropout=config.dropout,
                        dt_rank=config.dt_rank,
                        d_state=config.d_state,
                        dt_scale=config.dt_scale,
                        dt_init=config.dt_init,
                        dt_max=config.dt_max,
                        dt_min=config.dt_min,
                        dt_init_floor=config.dt_init_floor,
                        norm=get_normalization_layer(config),  # type: ignore
                        activation=config.activation,
                        bidirectional=config.bidirectional,
                        use_learnable_interaction=config.use_learnable_interaction,
                        layer_norm_eps=config.layer_norm_eps,
                        AD_weight_decay=config.AD_weight_decay,
                        BC_layer_norm=config.BC_layer_norm,
                        use_pscan=config.use_pscan,
                    )
                )

        # Check the type of the last layer and append the desired one if necessary
        if config.last_layer == "attn":
            if not isinstance(self.layers[-1], nn.MultiheadAttention):
                self.layers.append(
                    nn.MultiheadAttention(
                        embed_dim=config.d_model,
                        num_heads=config.n_heads,
                        dropout=config.dropout,
                    )
                )
        else:
            if not isinstance(self.layers[-1], ResidualBlock):
                self.layers.append(
                    ResidualBlock(
                        d_model=config.d_model,
                        expand_factor=config.expand_factor,
                        bias=config.bias,
                        d_conv=config.d_conv,
                        conv_bias=config.conv_bias,
                        dropout=config.dropout,
                        dt_rank=config.dt_rank,
                        d_state=config.d_state,
                        dt_scale=config.dt_scale,
                        dt_init=config.dt_init,
                        dt_max=config.dt_max,
                        dt_min=config.dt_min,
                        dt_init_floor=config.dt_init_floor,
                        norm=get_normalization_layer(config),  # type: ignore
                        activation=config.activation,
                        bidirectional=config.bidirectional,
                        use_learnable_interaction=config.use_learnable_interaction,
                        layer_norm_eps=config.layer_norm_eps,
                        AD_weight_decay=config.AD_weight_decay,
                        BC_layer_norm=config.BC_layer_norm,
                        use_pscan=config.use_pscan,
                    )
                )

    def forward(self, x):
        for layer in self.layers:
            if isinstance(layer, nn.MultiheadAttention):
                # If it's an attention layer, handle input shape (seq_len, batch, embed_dim)
                # Switch to (seq_len, batch, embed_dim) for attention
                x = x.transpose(0, 1)
                x, _ = layer(x, x, x)
                # Switch back to (batch, seq_len, embed_dim)
                x = x.transpose(0, 1)
            else:
                # Otherwise, pass through Mamba block
                x = layer(x)

        return x



import torch
import numpy as np
from ..arch_utils.get_norm_fn import get_normalization_layer
from ..arch_utils.layer_utils.embedding_layer import EmbeddingLayer
from ..arch_utils.mamba_utils.mambattn_arch import MambAttn
from ..arch_utils.mlp_utils import MLPhead
from ..configs.mambattention_config import DefaultMambAttentionConfig
from .utils.basemodel import BaseModel


class MambAttention(BaseModel):
    """A MambAttention model for tabular data, integrating feature embeddings, attention-based Mamba transformations,
    and a customizable architecture for handling categorical and numerical features.

    Parameters
    ----------
    cat_feature_info : dict
        Dictionary containing information about categorical features, including their names and dimensions.
    num_feature_info : dict
        Dictionary containing information about numerical features, including their names and dimensions.
    num_classes : int, optional
        The number of output classes or target dimensions for regression, by default 1.
    config : DefaultMambAttentionConfig, optional
        Configuration object with model hyperparameters such as dropout rates, head layer sizes, attention settings,
        and other architectural configurations, by default DefaultMambAttentionConfig().
    **kwargs : dict
        Additional keyword arguments for the BaseModel class.

    Attributes
    ----------
    pooling_method : str
        Pooling method to aggregate features after the Mamba attention layer.
    shuffle_embeddings : bool
        Flag indicating if embeddings should be shuffled, as specified in the configuration.
    mamba : MambAttn
        Mamba attention layer to process embedded features.
    norm_f : nn.Module
        Normalization layer for the processed features.
    embedding_layer : EmbeddingLayer
        Layer for embedding categorical and numerical features.
    tabular_head : MLPhead
        MLPhead layer to produce the final prediction based on the output of the Mamba attention layer.
    perm : torch.Tensor, optional
        Permutation tensor used for shuffling embeddings, if enabled.

    Methods
    -------
    forward(num_features, cat_features)
        Perform a forward pass through the model, including embedding, Mamba attention transformation, pooling,
        and prediction steps.
    """

    def __init__(
        self,
        feature_information: tuple,  # Expecting (num_feature_info, cat_feature_info, embedding_feature_info)
        num_classes=1,
        config: DefaultMambAttentionConfig = DefaultMambAttentionConfig(),  # noqa: B008
        **kwargs,
    ):
        super().__init__(config=config, **kwargs)
        self.save_hyperparameters(ignore=["feature_information"])

        self.returns_ensemble = False

        try:
            self.pooling_method = self.hparams.pooling_method
        except AttributeError:
            self.pooling_method = config.pooling_method

        try:
            self.shuffle_embeddings = self.hparams.shuffle_embeddings
        except AttributeError:
            self.shuffle_embeddings = config.shuffle_embeddings

        self.mamba = MambAttn(config)
        self.norm_f = get_normalization_layer(config)

        # embedding layer
        self.embedding_layer = EmbeddingLayer(
            *feature_information,
            config=config,
        )

        try:
            head_activation = self.hparams.head_activation
        except AttributeError:
            head_activation = config.head_activation

        try:
            input_dim = self.hparams.d_model
        except AttributeError:
            input_dim = config.d_model

        self.tabular_head = MLPhead(
            input_dim=input_dim,
            config=config,
            output_dim=num_classes,
        )

        if self.shuffle_embeddings:
            self.perm = torch.randperm(self.embedding_layer.seq_len)

        # pooling
        n_inputs = np.sum([len(info) for info in feature_information])
        self.initialize_pooling_layers(config=config, n_inputs=n_inputs)

    def forward(self, *data):
        """Defines the forward pass of the model.

        Parameters
        ----------
        data : tuple
            Input tuple of tensors of num_features, cat_features, embeddings.

        Returns
        -------
        torch.Tensor
            Output tensor.
        """
        x = self.embedding_layer(*data)

        if self.shuffle_embeddings:
            x = x[:, self.perm, :]

        x = self.mamba(x)

        x = self.pool_sequence(x)

        x = self.norm_f(x)  # type: ignore
        preds = self.tabular_head(x)

        return preds



import pandas as pd
import numpy as np
import warnings


def check_inputs(
    X,
    y=None,
    numerical_columns=None,
    categorical_columns=None,
    task_type=None,
    min_samples=5,
):
    """
    Perform thorough validation on input features and target.

    Parameters
    ----------
    X : pd.DataFrame or dict
        Input features.
    y : array-like, optional
        Target values.
    numerical_columns : list of str
        Columns expected to be numerical.
    categorical_columns : list of str
        Columns expected to be categorical.
    task_type : str, optional
        One of {"regression", "binary", "multiclass"}. If specified, target checks will apply accordingly.
    min_samples : int, optional
        Minimum number of distinct values required in any feature or target.

    Raises
    ------
    ValueError
        If any feature or target fails validation checks.
    """
    if isinstance(X, dict):
        X = pd.DataFrame(X)

    if not isinstance(X, pd.DataFrame):
        raise TypeError("X must be a DataFrame or a dict convertible to DataFrame.")

    if X.empty:
        raise ValueError("X must not be empty.")

    if numerical_columns is None:
        numerical_columns = []
    if categorical_columns is None:
        categorical_columns = []

    all_cols = set(numerical_columns) | set(categorical_columns)
    missing_cols = all_cols - set(X.columns)
    if missing_cols:
        raise ValueError(
            f"The following specified columns are missing in X: {missing_cols}"
        )

    # Check numerical features
    for col in numerical_columns:
        series = X[col]
        if series.nunique(dropna=False) < min_samples:
            raise ValueError(
                f"Numerical feature '{col}' has less than {min_samples} unique values."
            )
        if not np.issubdtype(series.dtype, np.number):
            raise TypeError(f"Numerical feature '{col}' must be numeric.")
        if not np.all(np.isfinite(series.dropna())):
            raise ValueError(
                f"Numerical feature '{col}' contains non-finite values (inf or NaN)."
            )

    # Check categorical features
    for col in categorical_columns:
        series = X[col]
        if series.nunique(dropna=False) < 2:
            raise ValueError(
                f"Categorical feature '{col}' has less only a single value ."
            )
        if pd.api.types.is_numeric_dtype(
            series
        ) and not pd.api.types.is_categorical_dtype(series):
            # allow numerical dtypes only if user intends to encode them
            pass  # optionally warn or convert
        if series.isnull().all():
            raise ValueError(f"Categorical feature '{col}' contains only NaNs.")

    # Check y
    if y is not None:
        y = np.array(y)

        if y.ndim != 1:
            raise ValueError("y must be a 1D array or Series.")

        if len(y) != len(X):
            raise ValueError("X and y must have the same number of samples.")

        unique_targets = np.unique(y[~pd.isnull(y)])
        n_classes = len(unique_targets)

        if task_type == "regression":
            if not np.issubdtype(y.dtype, np.number):
                raise TypeError("For regression, target y must be numeric.")
            if not np.all(np.isfinite(y)):
                raise ValueError("Target y contains non-finite values.")

            if n_classes <= 10:
                warnings.warn(
                    f"Target y has only {n_classes} unique values. "
                    "Consider if this should be a classification problem instead of regression.",
                    UserWarning,
                )

        elif task_type == "classification":
            if n_classes < 2:
                raise ValueError("Classification tasks requires more than 1 class.")



import torch
import torch.nn as nn
from torch.nn.parameter import Parameter


class SNLinear(nn.Module):
    """Separate linear layers for each feature embedding."""

    def __init__(self, n: int, in_features: int, out_features: int) -> None:
        super().__init__()
        self.weight = Parameter(torch.empty(n, in_features, out_features))
        self.bias = Parameter(torch.empty(n, out_features))
        self.reset_parameters()

    def reset_parameters(self) -> None:
        d_in_rsqrt = self.weight.shape[-2] ** -0.5
        nn.init.uniform_(self.weight, -d_in_rsqrt, d_in_rsqrt)
        nn.init.uniform_(self.bias, -d_in_rsqrt, d_in_rsqrt)

    def forward(self, x):
        if x.ndim != 3:
            raise ValueError("SNLinear requires a 3D input (batch, features, embedding).")
        if x.shape[-(self.weight.ndim - 1) :] != self.weight.shape[:-1]:
            raise ValueError("Input shape mismatch with weight dimensions.")

        x = x.transpose(0, 1) @ self.weight
        return x.transpose(0, 1) + self.bias



from collections.abc import Callable
from dataclasses import dataclass, field
import torch.nn as nn
from .base_config import BaseConfig


@dataclass
class DefaultMambaTabConfig(BaseConfig):
    """Configuration class for the Default MambaTab model with predefined hyperparameters.

    Parameters
    ----------
    d_model : int, default=64
        Dimensionality of the model.
    n_layers : int, default=1
        Number of layers in the model.
    expand_factor : int, default=2
        Expansion factor for the feed-forward layers.
    bias : bool, default=False
        Whether to use bias in the linear layers.
    d_conv : int, default=16
        Dimensionality of the convolutional layers.
    conv_bias : bool, default=True
        Whether to use bias in the convolutional layers.
    dropout : float, default=0.05
        Dropout rate for regularization.
    dt_rank : str, default="auto"
        Rank of the decision tree used in the model.
    d_state : int, default=128
        Dimensionality of the state in recurrent layers.
    dt_scale : float, default=1.0
        Scaling factor for the decision tree.
    dt_init : str, default="random"
        Initialization method for the decision tree.
    dt_max : float, default=0.1
        Maximum value for decision tree initialization.
    dt_min : float, default=1e-04
        Minimum value for decision tree initialization.
    dt_init_floor : float, default=1e-04
        Floor value for decision tree initialization.
    activation : callable, default=nn.ReLU()
        Activation function for the model.
    axis : int, default=1
        Axis along which operations are applied, if applicable.
    head_layer_sizes : list, default=()
        Sizes of the fully connected layers in the model's head.
    head_dropout : float, default=0.0
        Dropout rate for the head layers.
    head_skip_layers : bool, default=False
        Whether to skip layers in the head.
    head_activation : callable, default=nn.ReLU()
        Activation function for the head layers.
    head_use_batch_norm : bool, default=False
        Whether to use batch normalization in the head layers.
    norm : str, default="LayerNorm"
        Type of normalization to be used ('LayerNorm', 'RMSNorm', etc.).
    use_pscan : bool, default=False
        Whether to use PSCAN for the state-space model.
    mamba_version : str, default="mamba-torch"
        Version of the Mamba model to use ('mamba-torch', 'mamba1', 'mamba2').
    bidirectional : bool, default=False
        Whether to process data bidirectionally.
    """

    # Architecture Parameters
    d_model: int = 64
    n_layers: int = 1
    expand_factor: int = 2
    bias: bool = False
    d_conv: int = 16
    conv_bias: bool = True
    dropout: float = 0.05
    dt_rank: str = "auto"
    d_state: int = 128
    dt_scale: float = 1.0
    dt_init: str = "random"
    dt_max: float = 0.1
    dt_min: float = 1e-04
    dt_init_floor: float = 1e-04
    activation: Callable = nn.ReLU()  # noqa: RUF009
    axis: int = 1

    # Head Parameters
    head_layer_sizes: list = field(default_factory=list)
    head_dropout: float = 0.0
    head_skip_layers: bool = False
    head_activation: Callable = nn.ReLU()  # noqa: RUF009
    head_use_batch_norm: bool = False

    # Additional Features
    norm: str = "LayerNorm"
    use_pscan: bool = False
    mamba_version: str = "mamba-torch"
    bidirectional: bool = False



from dataclasses import dataclass
from .base_config import BaseConfig


@dataclass
class DefaultNDTFConfig(BaseConfig):
    """Configuration class for the default Neural Decision Tree Forest (NDTF) model with predefined hyperparameters.

    Parameters
    ----------
    min_depth : int, default=2
        Minimum depth of trees in the forest. Controls the simplest model structure.
    max_depth : int, default=10
        Maximum depth of trees in the forest. Controls the maximum complexity of the trees.
    temperature : float, default=0.1
        Temperature parameter for softening the node decisions during path probability calculation.
    node_sampling : float, default=0.3
        Fraction of nodes sampled for regularization penalty calculation. Reduces computation by focusing
        on a subset of nodes.
    lamda : float, default=0.3
        Regularization parameter to control the complexity of the paths, penalizing overconfident
        or imbalanced paths.
    n_ensembles : int, default=12
        Number of trees in the forest
    penalty_factor : float, default=0.01
        Factor with which the penalty is multiplied
    """

    min_depth: int = 4
    max_depth: int = 16
    temperature: float = 0.1
    node_sampling: float = 0.3
    lamda: float = 0.3
    n_ensembles: int = 12
    penalty_factor: float = 1e-08



import numpy as np
import torch
import torch.distributions as dist


class BaseDistribution(torch.nn.Module):
    """
    The base class for various statistical distributions, providing a common interface and utilities.

    This class defines the basic structure and methods that are inherited by specific distribution
    classes, allowing for the implementation of custom distributions with specific parameter transformations
    and loss computations.

    Attributes
    ----------
        _name (str): The name of the distribution.
        param_names (list of str): A list of names for the parameters of the distribution.
        param_count (int): The number of parameters for the distribution.
        predefined_transforms (dict): A dictionary of predefined transformation functions for parameters.

    Parameters
    ----------
        name (str): The name of the distribution.
        param_names (list of str): A list of names for the parameters of the distribution.
    """

    def __init__(self, name, param_names):
        super().__init__()

        self._name = name
        self.param_names = param_names
        self.param_count = len(param_names)
        # Predefined transformation functions accessible to all subclasses
        self.predefined_transforms = {
            "positive": torch.nn.functional.softplus,
            "none": lambda x: x,
            "square": lambda x: x**2,
            "exp": torch.exp,
            "sqrt": torch.sqrt,
            "probabilities": lambda x: torch.softmax(x, dim=-1),
            # Adding a small constant for numerical stability
            "log": lambda x: torch.log(x + 1e-6),
        }

    @property
    def name(self):
        return self._name

    @property
    def parameter_count(self):
        return self.param_count

    def get_transform(self, transform_name):
        """
        Retrieve a transformation function by name, or return the function if it's custom.
        """
        if callable(transform_name):
            # Custom transformation function provided
            return transform_name
        # Default to 'none'
        return self.predefined_transforms.get(transform_name, lambda x: x)

    def compute_loss(self, predictions, y_true):
        """
        Computes the loss (e.g., negative log likelihood) for the distribution given
        predictions and true values.

        This method must be implemented by subclasses.

        Parameters
        ----------
            predictions (torch.Tensor): The predicted parameters of the distribution.
            y_true (torch.Tensor): The true values.

        Raises
        ------
            NotImplementedError: If the subclass does not implement this method.
        """
        raise NotImplementedError("Subclasses must implement this method.")

    def evaluate_nll(self, y_true, y_pred):
        """
        Evaluates the negative log likelihood (NLL) for given true values and predictions.

        Parameters
        ----------
            y_true (array-like): The true values.
            y_pred (array-like): The predicted values.

        Returns
        -------
            dict: A dictionary containing the NLL value.
        """

        # Convert numpy arrays to torch tensors
        y_true_tensor = torch.tensor(y_true, dtype=torch.float32)
        y_pred_tensor = torch.tensor(y_pred, dtype=torch.float32)

        # Compute NLL using the provided loss function
        nll_loss_tensor = self.compute_loss(y_pred_tensor, y_true_tensor)

        # Convert the NLL loss tensor back to a numpy array and return
        return {
            "NLL": nll_loss_tensor.detach().numpy(),
        }

    def forward(self, predictions):
        """
        Apply the appropriate transformations to the predicted parameters.

        Parameters:
            predictions (torch.Tensor): The predicted parameters of the distribution.

        Returns:
            torch.Tensor: A tensor with transformed parameters.
        """
        transformed_params = []
        for idx, param_name in enumerate(self.param_names):
            transform_func = self.get_transform(
                getattr(self, f"{param_name}_transform", "none")
            )
            transformed_params.append(
                transform_func(predictions[:, idx]).unsqueeze(  # type: ignore
                    1
                )  # type: ignore
            )
        return torch.cat(transformed_params, dim=1)


class NormalDistribution(BaseDistribution):
    """
    Represents a Normal (Gaussian) distribution with parameters for mean and variance,
    including functionality for transforming these parameters and computing the loss.

    Inherits from BaseDistribution.

    Parameters
    ----------
        name (str): The name of the distribution. Defaults to "Normal".
        mean_transform (str or callable): The transformation for the mean parameter.
        Defaults to "none".
        var_transform (str or callable): The transformation for the variance parameter.
        Defaults to "positive".
    """

    def __init__(self, name="Normal", mean_transform="none", var_transform="positive"):
        param_names = [
            "mean",
            "variance",
        ]
        super().__init__(name, param_names)

        self.mean_transform = self.get_transform(mean_transform)
        self.variance_transform = self.get_transform(var_transform)

    def compute_loss(self, predictions, y_true):
        mean = self.mean_transform(predictions[:, self.param_names.index("mean")])
        variance = self.variance_transform(
            predictions[:, self.param_names.index("variance")]
        )

        normal_dist = dist.Normal(mean, variance)

        nll = -normal_dist.log_prob(y_true).mean()
        return nll

    def evaluate_nll(self, y_true, y_pred):
        metrics = super().evaluate_nll(y_true, y_pred)

        # Convert numpy arrays to torch tensors
        y_true_tensor = torch.tensor(y_true, dtype=torch.float32)
        y_pred_tensor = torch.tensor(y_pred, dtype=torch.float32)

        mse_loss = torch.nn.functional.mse_loss(
            y_true_tensor, y_pred_tensor[:, self.param_names.index("mean")]
        )
        rmse = np.sqrt(mse_loss.detach().numpy())
        mae = (
            torch.nn.functional.l1_loss(
                y_true_tensor, y_pred_tensor[:, self.param_names.index("mean")]
            )
            .detach()
            .numpy()
        )

        metrics["mse"] = mse_loss.detach().numpy()
        metrics["mae"] = mae
        metrics["rmse"] = rmse

        # Convert the NLL loss tensor back to a numpy array and return
        return metrics


class PoissonDistribution(BaseDistribution):
    """
    Represents a Poisson distribution, typically used for modeling count data or the number of events
    occurring within a fixed interval of time or space. This class extends the BaseDistribution and
    includes parameter transformation and loss computation specific to the Poisson distribution.

    Parameters
    ----------
        name (str): The name of the distribution, defaulted to "Poisson".
        rate_transform (str or callable): Transformation to apply to the rate parameter
        to ensure it remains positive.
    """

    def __init__(self, name="Poisson", rate_transform="positive"):
        # Specify parameter name for Poisson distribution
        param_names = ["rate"]
        super().__init__(name, param_names)
        # Retrieve transformation function for rate
        self.rate_transform = self.get_transform(rate_transform)

    def compute_loss(self, predictions, y_true):
        rate = self.rate_transform(predictions[:, self.param_names.index("rate")])

        # Define the Poisson distribution with the transformed parameter
        poisson_dist = dist.Poisson(rate)

        # Compute the negative log-likelihood
        nll = -poisson_dist.log_prob(y_true).mean()
        return nll

    def evaluate_nll(self, y_true, y_pred):
        metrics = super().evaluate_nll(y_true, y_pred)

        # Convert numpy arrays to torch tensors
        y_true_tensor = torch.tensor(y_true, dtype=torch.float32)
        y_pred_tensor = torch.tensor(y_pred, dtype=torch.float32)
        rate = self.rate_transform(y_pred_tensor[:, self.param_names.index("rate")])

        mse_loss = torch.nn.functional.mse_loss(y_true_tensor, rate)  # type: ignore
        rmse = np.sqrt(mse_loss.detach().numpy())
        mae = (
            torch.nn.functional.l1_loss(y_true_tensor, rate)  # type: ignore
            .detach()
            .numpy()  # type: ignore
        )  # type: ignore
        poisson_deviance = 2 * torch.sum(
            y_true_tensor * torch.log(y_true_tensor / rate) - (y_true_tensor - rate)
        )

        metrics["mse"] = mse_loss.detach().numpy()
        metrics["mae"] = mae
        metrics["rmse"] = rmse
        metrics["poisson_deviance"] = poisson_deviance.detach().numpy()

        # Convert the NLL loss tensor back to a numpy array and return
        return metrics


class InverseGammaDistribution(BaseDistribution):
    """
    Represents an Inverse Gamma distribution, often used as a prior distribution in Bayesian statistics,
    especially for scale parameters in other distributions. This class extends BaseDistribution and includes
    parameter transformation and loss computation specific to the Inverse Gamma distribution.

    Parameters
    ----------
        name (str): The name of the distribution, defaulted to "InverseGamma".
        shape_transform (str or callable): Transformation for the shape parameter to
        ensure it remains positive.
        scale_transform (str or callable): Transformation for the scale parameter to
        ensure it remains positive.
    """

    def __init__(
        self,
        name="InverseGamma",
        shape_transform="positive",
        scale_transform="positive",
    ):
        param_names = [
            "shape",
            "scale",
        ]
        super().__init__(name, param_names)

        self.shape_transform = self.get_transform(shape_transform)
        self.scale_transform = self.get_transform(scale_transform)

    def compute_loss(self, predictions, y_true):
        shape = self.shape_transform(predictions[:, self.param_names.index("shape")])
        scale = self.scale_transform(predictions[:, self.param_names.index("scale")])

        inverse_gamma_dist = dist.InverseGamma(shape, scale)
        # Compute the negative log-likelihood
        nll = -inverse_gamma_dist.log_prob(y_true).mean()
        return nll


class BetaDistribution(BaseDistribution):
    """
    Represents a Beta distribution, a continuous distribution defined on the interval [0, 1], commonly used
    in Bayesian statistics for modeling probabilities. This class extends BaseDistribution and includes parameter
    transformation and loss computation specific to the Beta distribution.

    Parameters
    ----------
        name (str): The name of the distribution, defaulted to "Beta".
        shape_transform (str or callable): Transformation for the alpha (shape) parameter to ensure
        it remains positive.
        scale_transform (str or callable): Transformation for the beta (scale) parameter to ensure
        it remains positive.
    """

    def __init__(
        self,
        name="Beta",
        shape_transform="positive",
        scale_transform="positive",
    ):
        param_names = [
            "alpha",
            "beta",
        ]
        super().__init__(name, param_names)

        self.alpha_transform = self.get_transform(shape_transform)
        self.beta_transform = self.get_transform(scale_transform)

    def compute_loss(self, predictions, y_true):
        alpha = self.alpha_transform(predictions[:, self.param_names.index("alpha")])
        beta = self.beta_transform(predictions[:, self.param_names.index("beta")])

        beta_dist = dist.Beta(alpha, beta)
        # Compute the negative log-likelihood
        nll = -beta_dist.log_prob(y_true).mean()
        return nll


class DirichletDistribution(BaseDistribution):
    """
    Represents a Dirichlet distribution, a multivariate generalization of the Beta distribution. It is commonly
    used in Bayesian statistics for modeling multinomial distribution probabilities. This class extends
    BaseDistribution and includes parameter transformation and loss computation
    specific to the Dirichlet distribution.

    Parameters
    ----------
        name (str): The name of the distribution, defaulted to "Dirichlet".
        concentration_transform (str or callable): Transformation to apply to
        concentration parameters to ensure they remain positive.
    """

    def __init__(self, name="Dirichlet", concentration_transform="positive"):
        # For Dirichlet, param_names could be dynamically set based on the dimensionality of alpha
        # For simplicity, we're not specifying individual names for each concentration parameter
        param_names = ["concentration"]  # This is a simplification
        super().__init__(name, param_names)
        # Retrieve transformation function for concentration parameters
        self.concentration_transform = self.get_transform(concentration_transform)

    def compute_loss(self, predictions, y_true):
        # Apply the transformation to ensure all concentration parameters are positive
        # Assuming predictions is a 2D tensor where each row is a set of concentration parameters
        # for a Dirichlet distribution
        concentration = self.concentration_transform(predictions)

        dirichlet_dist = dist.Dirichlet(concentration)

        nll = -dirichlet_dist.log_prob(y_true).mean()
        return nll


class GammaDistribution(BaseDistribution):
    """
    Represents a Gamma distribution, a two-parameter family of continuous probability distributions. It's
    widely used in various fields of science for modeling a wide range of phenomena. This class extends
    BaseDistribution and includes parameter transformation and loss computation specific to
    the Gamma distribution.

    Parameters
    ----------
        name (str): The name of the distribution, defaulted to "Gamma".
        shape_transform (str or callable): Transformation for the shape parameter to ensure it remains positive.
        rate_transform (str or callable): Transformation for the rate parameter to ensure it remains positive.
    """

    def __init__(
        self, name="Gamma", shape_transform="positive", rate_transform="positive"
    ):
        param_names = ["shape", "rate"]
        super().__init__(name, param_names)

        self.shape_transform = self.get_transform(shape_transform)
        self.rate_transform = self.get_transform(rate_transform)

    def compute_loss(self, predictions, y_true):
        shape = self.shape_transform(predictions[:, self.param_names.index("shape")])
        rate = self.rate_transform(predictions[:, self.param_names.index("rate")])

        # Define the Gamma distribution with the transformed parameters
        gamma_dist = dist.Gamma(shape, rate)

        # Compute the negative log-likelihood
        nll = -gamma_dist.log_prob(y_true).mean()
        return nll


class StudentTDistribution(BaseDistribution):
    """
    Represents a Student's t-distribution, a family of continuous probability distributions that arise when
    estimating the mean of a normally distributed population in situations where the sample size is small.
    This class extends BaseDistribution and includes parameter transformation and loss computation specific
    to the Student's t-distribution.

    Parameters
    ----------
        name (str): The name of the distribution, defaulted to "StudentT".
        df_transform (str or callable): Transformation for the degrees of freedom parameter
        to ensure it remains positive.
        loc_transform (str or callable): Transformation for the location parameter.
        scale_transform (str or callable): Transformation for the scale parameter
        to ensure it remains positive.
    """

    def __init__(
        self,
        name="StudentT",
        df_transform="positive",
        loc_transform="none",
        scale_transform="positive",
    ):
        param_names = ["df", "loc", "scale"]
        super().__init__(name, param_names)

        self.df_transform = self.get_transform(df_transform)
        self.loc_transform = self.get_transform(loc_transform)
        self.scale_transform = self.get_transform(scale_transform)

    def compute_loss(self, predictions, y_true):
        df = self.df_transform(predictions[:, self.param_names.index("df")])
        loc = self.loc_transform(predictions[:, self.param_names.index("loc")])
        scale = self.scale_transform(predictions[:, self.param_names.index("scale")])

        student_t_dist = dist.StudentT(df, loc, scale)  # type: ignore

        nll = -student_t_dist.log_prob(y_true).mean()
        return nll

    def evaluate_nll(self, y_true, y_pred):
        metrics = super().evaluate_nll(y_true, y_pred)

        # Convert numpy arrays to torch tensors
        y_true_tensor = torch.tensor(y_true, dtype=torch.float32)
        y_pred_tensor = torch.tensor(y_pred, dtype=torch.float32)

        mse_loss = torch.nn.functional.mse_loss(
            y_true_tensor, y_pred_tensor[:, self.param_names.index("loc")]
        )
        rmse = np.sqrt(mse_loss.detach().numpy())
        mae = (
            torch.nn.functional.l1_loss(
                y_true_tensor, y_pred_tensor[:, self.param_names.index("loc")]
            )
            .detach()
            .numpy()
        )

        metrics["mse"] = mse_loss.detach().numpy()
        metrics["mae"] = mae
        metrics["rmse"] = rmse

        # Convert the NLL loss tensor back to a numpy array and return
        return metrics


class NegativeBinomialDistribution(BaseDistribution):
    """
    Represents a Negative Binomial distribution, often used for count data and modeling the number
    of failures before a specified number of successes occurs in a series of Bernoulli trials.
    This class extends BaseDistribution and includes parameter transformation and loss computation
    specific to the Negative Binomial distribution.

    Parameters
    ----------
        name (str): The name of the distribution, defaulted to "NegativeBinomial".
        mean_transform (str or callable): Transformation for the mean parameter to ensure it remains positive.
        dispersion_transform (str or callable): Transformation for the dispersion parameter to
        ensure it remains positive.
    """

    def __init__(
        self,
        name="NegativeBinomial",
        mean_transform="positive",
        dispersion_transform="positive",
    ):
        param_names = ["mean", "dispersion"]
        super().__init__(name, param_names)

        self.mean_transform = self.get_transform(mean_transform)
        self.dispersion_transform = self.get_transform(dispersion_transform)

    def compute_loss(self, predictions, y_true):
        # Apply transformations to ensure mean and dispersion parameters are positive
        mean = self.mean_transform(predictions[:, self.param_names.index("mean")])
        dispersion = self.dispersion_transform(
            predictions[:, self.param_names.index("dispersion")]
        )

        # Calculate the probability (p) and number of successes (r) from mean and dispersion
        # These calculations follow from the mean and variance of the negative binomial distribution
        # where variance = mean + mean^2 / dispersion
        r = torch.tensor(1.0) / dispersion
        p = r / (r + mean)

        # Define the Negative Binomial distribution with the transformed parameters
        negative_binomial_dist = dist.NegativeBinomial(total_count=r, probs=p)

        # Compute the negative log-likelihood
        nll = -negative_binomial_dist.log_prob(y_true).mean()
        return nll


class CategoricalDistribution(BaseDistribution):
    """
    Represents a Categorical distribution, a discrete distribution that describes the possible results of a
    random variable that can take on one of K possible categories, with the probability of each category
    separately specified. This class extends BaseDistribution and includes parameter transformation and loss
    computation specific to the Categorical distribution.

    Parameters
    ----------
        name (str): The name of the distribution, defaulted to "Categorical".
        prob_transform (str or callable): Transformation for the probabilities to ensure
        they remain valid (i.e., non-negative and sum to 1).
    """

    def __init__(self, name="Categorical", prob_transform="probabilities"):
        # Specify parameter name for Poisson distribution
        param_names = ["probs"]
        super().__init__(name, param_names)
        # Retrieve transformation function for rate
        self.probs_transform = self.get_transform(prob_transform)

    def compute_loss(self, predictions, y_true):
        probs = self.probs_transform(predictions)

        # Define the Poisson distribution with the transformed parameter
        cat_dist = dist.Categorical(probs=probs)

        # Compute the negative log-likelihood
        nll = -cat_dist.log_prob(y_true).mean()
        return nll


class Quantile(BaseDistribution):
    """
    Quantile Regression Loss class.

    This class computes the quantile loss (also known as pinball loss) for a set of quantiles.
    It is used to handle quantile regression tasks where we aim to predict a given quantile of the target distribution.

    Parameters
    ----------
    name : str, optional
        The name of the distribution, by default "Quantile".
    quantiles : list of float, optional
        A list of quantiles to be used for computing the loss, by default [0.25, 0.5, 0.75].

    Attributes
    ----------
    quantiles : list of float
        List of quantiles for which the pinball loss is computed.

    Methods
    -------
    compute_loss(predictions, y_true)
        Computes the quantile regression loss between the predictions and true values.
    """

    def __init__(self, name="Quantile", quantiles=[0.25, 0.5, 0.75]):
        # Use string representations of quantiles
        param_names = [f"q_{q}" for q in quantiles]
        super().__init__(name, param_names)
        self.quantiles = quantiles

    def compute_loss(self, predictions, y_true):
        if y_true.requires_grad:
            raise ValueError("y_true should not require gradients")
        if predictions.size(0) != y_true.size(0):
            raise ValueError("Batch size of predictions and y_true must match")

        losses = []
        for i, q in enumerate(self.quantiles):
            # Calculate errors for each quantile
            errors = y_true - predictions[:, i]
            # Compute the pinball loss
            quantile_loss = torch.max((q - 1) * errors, q * errors)
            losses.append(quantile_loss)

        # Sum losses across quantiles and compute mean
        loss = torch.mean(torch.stack(losses, dim=1).sum(dim=1))
        return loss


class JohnsonSuDistribution(BaseDistribution):
    """
    Represents a Johnson's SU distribution with parameters for skewness, shape, location, and scale.

    Parameters
    ----------
        name (str): The name of the distribution. Defaults to "JohnsonSu".
        skew_transform (str or callable): The transformation for the skewness parameter. Defaults to "none".
        shape_transform (str or callable): The transformation for the shape parameter. Defaults to "positive".
        loc_transform (str or callable): The transformation for the location parameter. Defaults to "none".
        scale_transform (str or callable): The transformation for the scale parameter. Defaults to "positive".
    """

    def __init__(
        self,
        name="JohnsonSu",
        skew_transform="none",
        shape_transform="positive",
        loc_transform="none",
        scale_transform="positive",
    ):
        param_names = ["skew", "shape", "location", "scale"]
        super().__init__(name, param_names)

        self.skew_transform = self.get_transform(skew_transform)
        self.shape_transform = self.get_transform(shape_transform)
        self.loc_transform = self.get_transform(loc_transform)
        self.scale_transform = self.get_transform(scale_transform)

    def log_prob(self, x, skew, shape, loc, scale):
        """
        Compute the log probability density of the Johnson's SU distribution.
        """
        z = skew + shape * torch.asinh((x - loc) / scale)
        log_pdf = (
            torch.log(shape / (scale * np.sqrt(2 * np.pi)))
            - 0.5 * z**2
            - 0.5 * torch.log(1 + ((x - loc) / scale) ** 2)
        )
        return log_pdf

    def compute_loss(self, predictions, y_true):
        skew = self.skew_transform(predictions[:, self.param_names.index("skew")])
        shape = self.shape_transform(predictions[:, self.param_names.index("shape")])
        loc = self.loc_transform(predictions[:, self.param_names.index("location")])
        scale = self.scale_transform(predictions[:, self.param_names.index("scale")])

        log_probs = self.log_prob(y_true, skew, shape, loc, scale)
        nll = -log_probs.mean()
        return nll

    def evaluate_nll(self, y_true, y_pred):
        metrics = super().evaluate_nll(y_true, y_pred)

        y_true_tensor = torch.tensor(y_true, dtype=torch.float32)
        y_pred_tensor = torch.tensor(y_pred, dtype=torch.float32)

        mse_loss = torch.nn.functional.mse_loss(
            y_true_tensor, y_pred_tensor[:, self.param_names.index("location")]
        )
        rmse = np.sqrt(mse_loss.detach().numpy())
        mae = (
            torch.nn.functional.l1_loss(
                y_true_tensor, y_pred_tensor[:, self.param_names.index("location")]
            )
            .detach()
            .numpy()
        )

        metrics.update({"mse": mse_loss.detach().numpy(), "mae": mae, "rmse": rmse})

        return metrics
